from typing import Optional, List, Dict, Any
from ai_dashboard.tabs.data_report.blocks.base import BaseBlock


def _convert_to_bulk_editor_filters(filters: List[Dict[str, Any]]):
    filters = [{"filter_type": "or", "condition_value": filter} for filter in filters]
    filters[-1]["group"] = "end"
    return filters


class ConnectedChart(BaseBlock):
    def __init__(
        self,
        groupby: List[Dict[str, Any]],
        metric: List[Dict[str, Any]],
        chart_type: str = "column",
        title: Optional[str] = None,
        page_size: int = 20,
        show_frequencies: bool = False,
        sentiment_field: str = "",
        y_axis_sort_field: str = "",
        sort_direction: str = "desc",
        filters: Optional[List[Dict[str, Any]]] = None,
        convert_filters: bool = True,
    ):
        """
        Example for groupby
        [
            {
                "agg": "max",
                "field": "score",
                "color": "",
                "name": "max thumbsUpCount",
                "lowerIsBetter": false
            }
        ]
        Example for metric
        [
            {
                "agg": "category",
                "field": "content"
            }
        ]
        """
        assert sort_direction in {"desc", "asc"}

        for query in groupby:
            query["aggType"] = "groupby"

        for query in metric:
            query["aggType"] = "metric"

        if filters is None:
            filters = []

        if filters and convert_filters:
            filters = _convert_to_bulk_editor_filters(filters)

        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "datasetAggregation",
                    "attrs": {
                        "uid": "",
                        "title": title,
                        "chartType": chart_type,
                        "filters": filters,
                        "xAxis": {
                            "fields": groupby,
                            "numResults": page_size,
                            "resortAlphanumerically": False,
                        },
                        "yAxis": {
                            "fields": metric,
                            "showFrequency": show_frequencies,
                            "sortBy": y_axis_sort_field,
                            "sortDirection": sort_direction,
                        },
                        "timeseries": {
                            "field": "insert_date_",
                            "interval": "monthly",
                        },
                        "sentiment": {
                            "field": sentiment_field,
                            "mode": "overview",
                            "interval": "monthly",
                        },
                        "wordCloud": {"mode": "cloud"},
                    },
                }
            ],
        }


class ConnectedMetric(BaseBlock):
    def __init__(
        self,
        query: List[Dict[str, Any]],
        title: Optional[str] = None,
        show_frequencies: bool = False,
        sort_direction: str = "Descending",
    ):
        """
        Example for aggregates
        [
            {
                "agg": "category",
                "field": "_cluster_.desc_all-mpnet-base-v2_vector_.kmeans-8",
                "name": "category _cluster_.desc_all-mpnet-base-v2_vector_.kmeans-8",
                "aggType": "groupby",
            },
            {
                "agg": "avg",
                "field": "_sentiment_.desc.cardiffnlp-twitter-roberta-base-sentiment.overall_sentiment_score",
                "name": "avg desc (Sentiment Score)",
                "aggType": "metric",
            },
        ]
        """
        assert sort_direction in {"Descending", "Ascending"}

        for q in query:
            q["aggType"] = "metric"

        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "metricAggregation",
                    "attrs": {
                        "displayType": "column",
                        "sortDirection": sort_direction,
                        "showFrequencies": show_frequencies,
                        "datasetId": "cat_and_dogs",
                        "aggregates": query,
                        "sortBy": "",
                        "title": "" if title is None else title,
                    },
                }
            ],
        }



"""Marks for accenting basic text blocks"""
from ai_dashboard.tabs.data_report.blocks import BaseMark


class BoldMark(BaseMark):
    def __init__(self, content):
        self.mark = [{"type": "bold"}]
        self.block = self.preprocess_mark(content, self.mark)


class ItalicMark(BaseMark):
    def __init__(self, content):
        self.mark = [{"type": "italic"}]
        self.block = self.preprocess_mark(content, self.mark)


class StrikeMark(BaseMark):
    def __init__(self, content):
        self.mark = [{"type": "strike"}]
        self.block = self.preprocess_mark(content, self.mark)


class UnderlineMark(BaseMark):
    def __init__(self, content):
        self.mark = [{"type": "underline"}]
        self.block = self.preprocess_mark(content, self.mark)


class InlineCodeMark(BaseMark):
    def __init__(self, content):
        self.mark = [{"type": "code"}]
        self.block = self.preprocess_mark(content, self.mark)


class HighlightMark(BaseMark):
    def __init__(self, content, start, end, color):
        self.mark = [
            {
                "type": "highlight",
                "attrs": {"color": color},
                "from": start,
                "to": end,
            }
        ]
        self.block = self.preprocess_mark(content, self.mark)


class ColorMark(BaseMark):
    def __init__(self, content, color, background_color):
        self.mark = [
            {
                "type": "textStyle",
                "attrs": {"color": color, "backgroundColor": background_color},
            }
        ]
        self.block = self.preprocess_mark(content, self.mark)


class LinkMark(BaseMark):
    def __init__(self, content, href):
        self.mark = [{"type": "link", "attrs": {"href": href, "target": "_blank"}}]
        self.block = self.preprocess_mark(content, self.mark)



from ai_dashboard.tabs.data_report.blocks.base import BaseBlock


class H1(BaseBlock):
    def __init__(self, content):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "heading",
                    "attrs": {"level": 1},
                    "content": self.preprocess_text_contents(content),
                }
            ],
        }


class H2(BaseBlock):
    def __init__(self, content):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "heading",
                    "attrs": {"level": 2},
                    "content": self.preprocess_text_contents(content),
                }
            ],
        }


class H3(BaseBlock):
    def __init__(self, content):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "heading",
                    "attrs": {"level": 3},
                    "content": self.preprocess_text_contents(content),
                }
            ],
        }


class Paragraph(BaseBlock):
    def __init__(self, content: str):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {"type": "paragraph", "content": self.preprocess_text_contents(content)}
            ],
        }


class Code(BaseBlock):
    def __init__(self, text, language="python"):
        if not isinstance(text, str):
            raise TypeError("Code only accepts text type as input")
        # has to be text block.
        self.block = {
            "type": "codeBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "attrs": {"language": language},
            "content": [{"type": "text", "text": text}],
        }


P = Paragraph


class Space(BaseBlock):
    def __init__(self, height: int = 40):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {"type": "spaceBlock", "attrs": {"width": "100%", "height": height}}
            ],
        }



import os
import time
import warnings

from typing import Optional, Dict, Any

from ai_dashboard.api import endpoints, helpers
from ai_dashboard.dashboard import dashboard

from ai_dashboard import constants
from ai_dashboard import errors


class Client:
    def __init__(self, token: str = None, authenticate: bool = True) -> None:
        if token is None:
            token = os.getenv("DEVELOPMENT_TOKEN")

        self._credentials = helpers.process_token(token)
        self._token = token
        self._endpoints = endpoints.Endpoints(credentials=self._credentials)

        if authenticate:
            try:
                response = self.list_deployables()
            except:
                raise errors.AuthException
            else:
                self._deployables = sorted(
                    response["deployables"], key=lambda x: x["insert_date_"]
                )
                print(constants.WELCOME_MESSAGE.format(self._credentials.project))
        else:
            warnings.warn(
                "You have opted to not authenticate on client instantiation. Your token may or may not be valid."
            )

    @property
    def deployables(self):
        response = self.list_deployables()
        self._deployables = sorted(
            response["deployables"], key=lambda x: x["insert_date_"]
        )
        return self._deployables

    def _find_config(
        self,
        dataset_id: str = None,
        project_id: str = None,
        deployable_name: str = None,
    ):
        for deployable_json in self.deployables:
            if dataset_id is not None:
                dataset_id_match = deployable_json.get("dataset_id") == dataset_id
            else:
                dataset_id_match = True

            if project_id is not None:
                project_id_match = deployable_json.get("project_id") == project_id
            else:
                project_id_match = True

            if deployable_name is not None:
                deployable_config = deployable_json.get("configuration", {})
                deployable_name_match = (
                    deployable_config.get("deployable_name") == deployable_name
                )
            else:
                deployable_name = True

            if dataset_id_match and deployable_name_match and project_id_match:
                dataset_id = deployable_json["dataset_id"]
                project_id = deployable_json["project_id"]
                deployable_id = deployable_json["deployable_id"]
                return dataset_id, project_id, deployable_id, deployable_config

        raise ValueError(
            f"No deployable found that matches `dataset_id`={dataset_id} and/or `project_id`={project_id} and/or `deployablable_name`={deployable_name}"
        )

    def load_dashboard(
        self,
        deployable_name: str = None,
        dataset_id: str = None,
        project_id: str = None,
    ) -> dashboard.Dashboard:
        dataset_id, deployable_id, project_id, configuration = self._find_config(
            dataset_id=dataset_id,
            project_id=project_id,
            deployable_name=deployable_name,
        )
        return dashboard.Dashboard(
            endpoints=self._endpoints,
            dataset_id=dataset_id,
            deployable_id=deployable_id,
            project_id=project_id,
            configuration=configuration,
        )

    def recent(self) -> dashboard.Dashboard:
        try:
            configuration = self.deployables[-1]
        except IndexError:
            raise IndexError("You have no dashboards")
        else:
            dataset_id = configuration.get("dataset_id")
            deployable_id = configuration.get("deployable_id")
            project_id = configuration.get("project_id")
            return dashboard.Dashboard(
                endpoints=self._endpoints,
                dataset_id=dataset_id,
                deployable_id=deployable_id,
                configuration=configuration["configuration"],
                project_id=project_id,
            )

    def list_deployables(self, page_size: int = 1000):
        return self._endpoints._list_deployables(page_size=page_size)

    def create_deployable(
        self,
        dataset_id: Optional[str] = None,
        deployable_id: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
    ) -> Dict[str, Any]:
        if deployable_id is None:
            if name is not None:
                configs = [
                    d
                    for d in self.deployables
                    if d["configuration"].get("deployable_name", "") == name
                    and d.get("dataset_id") == dataset_id
                ]
                try:
                    deployable_id = configs[0]["deployable_id"]
                except:
                    pass
                else:
                    return self._endpoints._get_deployable(
                        deployable_id=deployable_id,
                    )

            if config is None:
                config = {
                    "private": False,
                    "tabs": {},
                    "filters": {},
                    "dataset_name": dataset_id,
                    "type": "explorer",
                    "filterToggles": {},
                    "deployable_name": name,
                    "cacheTimestamp": int(time.time() * 1000),
                    "api_key": "",
                    "project_id": self._credentials.project,
                    "tabOrder": [],
                    "deployable_id": "",
                    "read_key": "",
                }
            return self._endpoints._create_deployable(
                dataset_id=dataset_id,
                config=config,
            )
        else:
            return self._endpoints._get_deployable(
                deployable_id=deployable_id,
            )

    def delete_deployable(self, deployable_id: str) -> None:
        return self._endpoints._delete_deployable(deployable_id=deployable_id)

    def Dashboard(
        self,
        dataset_id: Optional[str] = None,
        deployable_id: Optional[str] = None,
        name: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
    ) -> dashboard.Dashboard:
        assert (
            dataset_id is not None or deployable_id is not None
        ), "Set either `dataset_id` and `config` or `deployable_id`"
        response = self.create_deployable(
            dataset_id=dataset_id,
            deployable_id=deployable_id,
            config=config,
            name=name,
        )
        return dashboard.Dashboard(endpoints=self._endpoints, **response)

    def insert_temp_local_media(self, file_path: str):
        """
        Insert temporary local media.
        """
        data = self._endpoints._get_temp_file_upload_url()
        upload_url = data["upload_url"]
        download_url = data["download_url"]
        with open(file_path, "rb") as fn_byte:
            media_content = bytes(fn_byte.read())
        self._endpoints._upload_temporary_media(
            presigned_url=upload_url,
            media_content=media_content,
        )
        return {"download_url": download_url}



import io
import json
import plotly
import numpy as np

from typing import Any

from ai_dashboard.tabs.data_report.blocks.base import BaseBlock


class Plotly(BaseBlock):
    def __init__(
        self,
        fig: Any,
        title: str = "",
        width: int = None,
        height: int = None,
        width_percentage: int = 100,
        options=None,
    ):
        if options is None:
            options = {"displayLogo": False}
        try:
            import plotly
        except ImportError:
            raise ImportError(
                ".plotly requires plotly to be installed, install with 'pip install -U plotly'."
            )
        layout = fig._layout
        if width:
            layout["width"] = width
        else:
            if "width" in layout and layout["width"] == np.inf:
                layout["width"] = "auto"
        if height:
            layout["height"] = height
        else:
            if "height" in layout and layout["height"] == np.inf:
                layout["height"] = "auto"
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "attrs": {
                        "height": "auto",
                        "layout": layout,
                        "options": options,
                        "title": title,
                        "width": f"{width_percentage}%",
                        **json.loads(plotly.io.to_json(fig)),
                    },
                    "type": "plotlyChart",
                }
            ],
        }


class Altair(BaseBlock):
    def __init__(
        self,
        fig: Any,
        title: str = "",
        width_percentage: int = 100,
    ):
        try:
            import altair
        except ImportError:
            raise ImportError(
                ".altair requires altair to be installed, install with 'pip install -U altair'."
            )
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "attrs": {
                        "height": "auto",
                        "title": title,
                        "width": f"{width_percentage}%",
                        "spec": json.loads(fig.to_json()),
                    },
                    "type": "vegaChart",
                }
            ],
        }


class Vega(BaseBlock):
    def __init__(
        self,
        vega_json: Any,
        title: str = "",
        width_percentage: int = 100,
    ):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "attrs": {
                        "height": "auto",
                        "title": title,
                        "width": f"{width_percentage}%",
                        "spec": vega_json,
                    },
                    "type": "vegaChart",
                }
            ],
        }


class Pyplot(BaseBlock):
    def __init__(
        self,
        fig,
        title: str = "",
        width: int = None,
        height: int = None,
        dpi: int = 150,
        width_percentage: int = 100,
        **kwargs,
    ):
        self.title = title
        self.width_percentage = width_percentage
        self.upload = True
        try:
            import matplotlib.pyplot as plt

            plt.ioff()
        except ImportError:
            raise ImportError(
                ".pyplot requires matplotlib to be installed, install with 'pip install -U matplotlib'."
            )
        if width or height:
            sizes = fig.get_size_inches()
            fig.set_size_inches(
                width / dpi if width else sizes[0], height / dpi if height else sizes[1]
            )
        self.fig_image = io.BytesIO()
        savefig_kwargs = {"dpi": dpi, "bbox_inches": "tight", "format": "png"}
        savefig_kwargs.update(kwargs)
        fig.savefig(self.fig_image, **savefig_kwargs)

    def override_image_url(self, image_url: str):
        self.image_url = image_url
        self.upload = False

    def postprocess(self, report_client):
        """
        Upload the image to the report app and return the block.
        """
        if self.upload:
            self.image_url = report_client._endpoints._upload_file(
                report_client.dataset_id, self.fig_image, f"{self.title}.png"
            )
        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "imageDisplayBlock",
                    "attrs": {
                        "imageSrc": self.image_url,
                        "title": self.title,
                        "width": f"{self.width_percentage}%",
                        "height": "auto",
                    },
                }
            ],
        }



import uuid
from ai_dashboard.tabs.data_report.blocks.base import BaseBlock


class Image(BaseBlock):
    def __init__(
        self,
        image_url: str,
        title: str = "",
        width_percentage: int = 100,
        upload: bool = False,
    ):
        self.title = title
        self.image_url = image_url
        self.upload = upload
        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "imageDisplayBlock",
                    "attrs": {
                        "imageSrc": image_url,
                        "title": title,
                        "width": f"{width_percentage}%",
                        "height": "auto",
                    },
                }
            ],
        }

    def postprocess(self, report_client):
        """
        Upload the image to the report app and return the block.
        """
        if "https://userdata" in self.image_url and not self.upload:
            self.file_url = self.image_url
        else:
            filename = f"{self.title}.png" if self.title else f"{str(uuid.uuid4())}.png"
            self.file_url = report_client._endpoints._upload_file(
                report_client.dataset_id, self.image_url, filename
            )
        self.block["content"][0]["attrs"]["imageSrc"] = self.file_url


class Audio(BaseBlock):
    def __init__(self, audio_url: str, title: str = "", width_percentage: int = 100):
        self.title = title
        self.audio_url = audio_url
        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "audioDisplayBlock",
                    "attrs": {
                        "audioSrc": audio_url,
                        "title": title,
                        "width": f"{width_percentage}%",
                        "height": "auto",
                    },
                }
            ],
        }

    def postprocess(self, report_client):
        """
        Upload the image to the report app and return the block.
        """
        if "https://userdata" in self.audio_url and not self.upload:
            self.file_url = self.audio_url
        else:
            filename = f"{self.title}.wav" if self.title else f"{str(uuid.uuid4())}.wav"
            self.file_url = report_client._endpoints._upload_file(
                report_client.dataset_id, self.audio_url, filename
            )
        self.block["content"][0]["attrs"]["audioSrc"] = self.file_url


class Video(BaseBlock):
    def __init__(self, video_url: str, title: str = "", width_percentage: int = 100):
        self.title = title
        self.video_url = video_url
        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "videoDisplayBlock",
                    "attrs": {
                        "videoSrc": video_url,
                        "title": title,
                        "width": f"{width_percentage}%",
                        "height": "auto",
                    },
                }
            ],
        }

    def postprocess(self, report_client):
        """
        Upload the image to the report app and return the block.
        """
        if "https://userdata" in self.video_url and not self.upload:
            self.file_url = self.video_url
        else:
            filename = f"{self.title}.mp4" if self.title else f"{str(uuid.uuid4())}.mp4"
            self.file_url = report_client._endpoints._upload_file(
                report_client.dataset_id, self.video_url, filename
            )
        self.block["content"][0]["attrs"]["videoSrc"] = self.file_url


class PDF(BaseBlock):
    def __init__(
        self, pdf_url: str, title: str = "", width_percentage: int = 100, analysis={}
    ):
        self.title = title
        self.pdf_url = pdf_url
        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "pdfDisplayBlock",
                    "attrs": {
                        "pdfSrc": pdf_url,
                        "title": title,
                        "width": f"{width_percentage}%",
                        "height": "auto",
                    },
                }
            ],
        }
        if analysis:
            self.block["content"][0]["attrs"]["analysis"] = analysis

    def postprocess(self, report_client):
        """
        Upload the image to the report app and return the block.
        """
        if "https://userdata" in self.pdf_url and not self.upload:
            self.file_url = self.pdf_url
        else:
            filename = f"{self.title}.mp4" if self.title else f"{str(uuid.uuid4())}.mp4"
            self.file_url = report_client._endpoints._upload_file(
                report_client.dataset_id, self.pdf_url, filename
            )
        self.block["content"][0]["attrs"]["pdfSrc"] = self.file_url



"""Layout blocks can nest almost any other blocks"""
from ai_dashboard.tabs.data_report.blocks.base import BaseBlock


class ColumnContent(BaseBlock):
    def __init__(self, blocks):
        self.input_blocks = blocks
        self.block = {"type": "columnContent", "content": self.input_blocks}

    def postprocess(self, report_client=None):
        self.block = {
            "type": "columnContent",
            "content": self.postprocess_blocks(
                self.input_blocks, report_client=report_client
            ),
        }


class Columns(BaseBlock):
    def __init__(self, column_items, num_columns: int = 0):
        if not isinstance(column_items, list):
            raise TypeError("'column_items' needs to be a list")
        for c in column_items:
            if isinstance(c, dict):
                if "type" not in c or c["type"] != "columnContent":
                    raise TypeError(
                        "column in 'column_items' needs to be a ColumnContent"
                    )
            elif not isinstance(c, ColumnContent):
                print(type(c))
                raise TypeError("column in 'column_items' needs to be a ColumnContent")

        if num_columns:
            self.num_columns = num_columns
        else:
            self.num_columns = len(column_items)
        self.column_items = column_items
        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "columnBlock",
                    "attrs": {"columns": self.num_columns},
                    "content": self.column_items,
                }
            ],
        }

    def postprocess(self, report_client=None):
        self.block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "columnBlock",
                    "attrs": {"columns": self.num_columns},
                    "content": self.postprocess_blocks(self.column_items),
                }
            ],
        }


class Card(BaseBlock):
    def __init__(self, blocks, width, color):
        self.input_blocks = blocks
        self.width = width
        self.color = color
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "cardBlock",
                    "attrs": {"width": width, "color": color},
                    "content": blocks,
                }
            ],
        }

    def postprocess(self, report_client=None):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "cardBlock",
                    "attrs": {"width": self.width, "color": self.color},
                    "content": self.postprocess_blocks(
                        self.input_blocks, report_client=report_client
                    ),
                }
            ],
        }


class Tooltip(BaseBlock):
    def __init__(self, blocks, tooltip_text):
        self.tooltip_text = tooltip_text
        self.input_blocks = blocks
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "tooltip",
                    "attrs": {"content": tooltip_text},
                    "content": self.input_blocks,
                }
            ],
        }

    def postprocess(self, report_client=None):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "tooltip",
                    "attrs": {"content": self.tooltip_text},
                    "content": self.postprocess_blocks(
                        self.input_blocks, report_client=report_client
                    ),
                }
            ],
        }


class Quote(BaseBlock):
    def __init__(self, blocks):
        self.input_blocks = blocks
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [{"type": "blockquote", "content": blocks}],
        }

    def postprocess(self, report_client=None):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "blockquote",
                    "content": self.postprocess_blocks(
                        self.input_blocks, report_client=report_client
                    ),
                }
            ],
        }


class Details(BaseBlock):
    # This doesnt work in the frontend?
    def __init__(self, blocks, title_content, collapsed: bool = True):
        self.collapsed = collapsed
        self.input_blocks = blocks
        self.title_content = title_content
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "details",
                    "attrs": {"open": self.collapsed},
                    "content": [
                        {
                            "type": "detailsSummary",
                            "content": self.preprocess_text_contents(title_content),
                        },
                        {
                            "type": "detailsContent",
                            "content": self.input_blocks,
                        },
                    ],
                }
            ],
        }

    def postprocess(self, report_client=None):
        self.block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "details",
                    "attrs": {"open": self.collapsed},
                    "content": [
                        {
                            "type": "detailsSummary",
                            "content": self.preprocess_text_contents(
                                self.title_content
                            ),
                        },
                        {
                            "type": "detailsContent",
                            "content": self.postprocess_blocks(
                                self.input_blocks, report_client=report_client
                            ),
                        },
                    ],
                }
            ],
        }



import requests
import io

from ai_dashboard import __version__
from ai_dashboard.types import Credentials

from typing import Optional, Dict, Any, List


def json_decode(request):
    try:
        decoded = request.json()
    except:
        decoded = request.content
    return decoded


class Endpoints:
    def __init__(self, credentials: Credentials) -> None:
        self._credentials = credentials
        self._base_url = (
            f"https://api-{self._credentials.region}.stack.tryrelevance.com/latest"
        )
        self._headers = dict(
            Authorization=f"{self._credentials.project}:{self._credentials.api_key}",
        )

    def _create_deployable(
        self, dataset_id: Optional[str] = None, config: Optional[Dict[str, Any]] = None
    ):
        return json_decode(
            requests.post(
                url=self._base_url + "/deployables/create",
                headers=self._headers,
                json=dict(
                    dataset_id=dataset_id,
                    configuration={} if config is None else config,
                ),
            )
        )

    def _share_deployable(self, deployable_id: str):
        return json_decode(
            requests.post(
                url=self._base_url + f"/deployables/{deployable_id}/share",
                headers=self._headers,
            )
        )

    def _unshare_deployable(self, deployable_id: str):
        return json_decode(
            requests.post(
                url=self._base_url + f"/deployables/{deployable_id}/private",
                headers=self._headers,
            )
        )

    def _update_deployable(
        self,
        deployable_id: str,
        dataset_id: str,
        configuration: Optional[Dict] = None,
        overwrite: bool = True,
        upsert: bool = True,
    ):
        return json_decode(
            requests.post(
                url=self._base_url + f"/deployables/{deployable_id}/update",
                headers=self._headers,
                json=dict(
                    dataset_id=dataset_id,
                    configuration=configuration,
                    overwrite=overwrite,
                    upsert=upsert,
                ),
            )
        )

    def _share_dashboard(self, deployable_id: str):
        return json_decode(
            requests.post(
                url=self._base_url + f"/deployablegroups/{deployable_id}/share",
                headers=self._headers,
            )
        )

    def _unshare_dashboard(self, deployable_id: str):
        return json_decode(
            requests.post(
                url=self._base_url + f"/deployablegroups/{deployable_id}/private",
                headers=self._headers,
            )
        )

    def _get_deployable(self, deployable_id: str):
        return json_decode(
            requests.get(
                url=self._base_url + f"/deployables/{deployable_id}/get",
                headers=self._headers,
            )
        )

    def _delete_deployable(self, deployable_id: str):
        return json_decode(
            requests.post(
                url=self._base_url + f"/deployables/delete",
                headers=self._headers,
                json=dict(
                    id=deployable_id,
                ),
            )
        )

    def _list_deployables(self, page_size: int):
        return json_decode(
            requests.get(
                url=self._base_url + "/deployables/list",
                headers=self._headers,
                params=dict(
                    page_size=page_size,
                ),
            )
        )

    def _get_file_upload_urls(self, dataset_id: str, files: List[str]):
        return json_decode(
            requests.post(
                url=self._base_url + f"/datasets/{dataset_id}/get_file_upload_urls",
                headers=self._headers,
                json=dict(files=files),
            )
        )

    def _upload_media(self, presigned_url: str, media_content: bytes):
        if not isinstance(media_content, bytes):
            raise ValueError(
                f"media needs to be in a bytes format. Currently in {type(media_content)}"
            )
        return requests.put(
            presigned_url,
            data=media_content,
        )

    ##Below are the functions that are not part of the API
    def _get_content_bytes(self, content):
        if isinstance(content, str):
            if "http" in content and "/":
                # online image
                content_bytes = io.BytesIO(requests.get(content).content).getvalue()
            else:
                # local filepath
                content_bytes = io.BytesIO(open(content, "rb").read()).getvalue()
        elif isinstance(content, bytes):
            content_bytes = content
        elif isinstance(content, io.BytesIO):
            content_bytes = content.getvalue()
        else:
            raise TypeError("'content' needs to be of type str, bytes or io.BytesIO.")
        return content_bytes

    def _upload_file(self, dataset_id: str, data, filename: str):
        data_bytes = self._get_content_bytes(data)
        presigned_response = self._get_file_upload_urls(
            dataset_id=dataset_id, files=[filename]
        )
        response = self._upload_media(
            presigned_url=presigned_response["files"][0]["upload_url"],
            media_content=data_bytes,
        )
        assert response.status_code == 200
        return presigned_response["files"][0]["url"]

    def _upload_local_files(self, dataset_id: str, paths: List[str]):
        presigned_response = self._get_file_upload_urls(
            dataset_id=dataset_id, files=paths
        )
        urls = []
        for index, path in enumerate(paths):
            url = presigned_response["files"][index]["url"]
            upload_url = presigned_response["files"][index]["upload_url"]
            with open(path, "rb") as fn_byte:
                media_content = bytes(fn_byte.read())
            urls.append(url)
            response = self._upload_media(
                presigned_url=upload_url,
                media_content=media_content,
            )
            assert response.status_code == 200
        return urls

    def _get_temp_file_upload_url(self):
        """Use this for temporary file uploads.
        returns: {'download_url': ..., 'upload_url': ...}
        """
        response = requests.post(
            url=self._base_url + f"/services/get_temporary_file_upload_url",
            headers=self._headers,
        )
        return json_decode(response)

    def _upload_temporary_media(self, presigned_url: str, media_content: bytes):
        return requests.put(
            presigned_url, headers={"x-amz-tagging": "Expire=true"}, data=media_content
        )



from typing import Optional, Dict, Any, List

from ai_dashboard.api import endpoints
from ai_dashboard.tabs import abstract_tab


class CategoryView(abstract_tab.Tab):
    ID = "CLUSTER_VIEW"

    BLANK: Dict[str, Any] = {
        "activeFilterGroup": "",
        "color": None,
        "configuration": {},
        "name": "",
        "type": ID,
    }

    def __init__(
        self,
        endpoints: endpoints.Endpoints,
        dataset_id: str,
        title: Optional[str] = None,
        name: Optional[str] = None,
        colour: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()

        self._endpoints = endpoints

        if config is not None:
            self._config = config
        else:
            self.reset()

        self.dataset_id = dataset_id
        self.config["colour"] = colour
        self.config["configuration"]["title"] = self.ID if title is None else title
        self.config["name"] = name

    @classmethod
    def from_fields(
        cls,
        cluster_field: str,
        primary_field: str = None,
        secondary_field: str = None,
        other_metadata: List[str] = None,
        subcluster_field: str = "",
        layout_template: str = "custom",
        title: Optional[str] = None,
        name: Optional[str] = None,
        colour: Optional[str] = None,
        sort_by_metric: str = "frequency",
        overview_chart_type: str = "grid",
        sentiment_field: str = "",
        timeseries_interval: str = "monthly",
        timeseries_field: str = "insert_date_",
        metrics: list = None,
        sort_by_direction: str = "desc",
        results_mode: str = "summary",
    ):
        category_view = cls(name=name, title=title, colour=colour)
        fields = []
        if primary_field is not None:
            fields.append(
                {
                    "id": "primary",
                    "value": primary_field,
                },
            )
        if secondary_field is not None:
            fields.append(
                {
                    "id": "secondary",
                    "value": secondary_field,
                }
            )
        if other_metadata is not None:
            fields.append(
                {
                    "id": "multiple-fields",
                    "value": other_metadata,
                }
            )
        category_view.config = {
            "activeFilterGroup": "",
            "color": colour,
            "configuration": {
                "previewDocumentCardConfiguration": {
                    "layoutTemplateConfiguration": {"fields": fields},
                    "layoutTemplate": layout_template,
                    "fields": fields,
                },
                "subclusterField": subcluster_field,
                "categoryMetrics": {
                    "sortByMetric": sort_by_metric,
                    "overviewChartType": overview_chart_type,
                    "sentimentField": sentiment_field,
                    "timeseriesInterval": timeseries_interval,
                    "timeseriesField": timeseries_field,
                    "metrics": [] if metrics is not None else metrics,
                    "sortByDirection": sort_by_direction,
                    "resultsMode": results_mode,
                },
                "clusterField": cluster_field,
            },
            "name": name,
            "type": cls.ID,
        }
        return category_view

    def set_view(
        self,
        cluster_field: str,
        primary_field: str = None,
        secondary_field: str = None,
        other_metadata: List[str] = None,
        subcluster_field: str = "",
        layout_template: str = "custom",
        sort_by_metric: str = "frequency",
        overview_chart_type: str = "grid",
        sentiment_field: str = "",
        timeseries_interval: str = "monthly",
        timeseries_field: str = "insert_date_",
        metrics: list = None,
        sort_by_direction: str = "desc",
        results_mode: str = "summary",
    ):
        self.config["configuration"] = {
            "previewDocumentCardConfiguration": {
                "layoutTemplateConfiguration": {"fields": []},
                "layoutTemplate": layout_template,
            },
            "subclusterField": subcluster_field,
            "clusterField": cluster_field,
        }
        if primary_field is not None:
            self.config["configuration"]["previewDocumentCardConfiguration"][
                "layoutTemplateConfiguration"
            ]["fields"].append(
                {
                    "id": "primary",
                    "value": primary_field,
                }
            )
        if secondary_field is not None:
            self.config["configuration"]["previewDocumentCardConfiguration"][
                "layoutTemplateConfiguration"
            ]["fields"].append(
                {
                    "id": "secondary",
                    "value": secondary_field,
                }
            )
        if other_metadata is not None:
            self.config["configuration"]["previewDocumentCardConfiguration"][
                "layoutTemplateConfiguration"
            ]["fields"].append(
                {
                    "id": "multiple-fields",
                    "value": other_metadata,
                }
            )
        self.config["configuration"]["categoryMetrics"] = {
            "sortByMetric": sort_by_metric,
            "overviewChartType": overview_chart_type,
            "sentimentField": sentiment_field,
            "timeseriesInterval": timeseries_interval,
            "timeseriesField": timeseries_field,
            "metrics": [] if metrics is not None else metrics,
            "sortByDirection": sort_by_direction,
            "resultsMode": results_mode,
        }



import io
import requests
import numpy as np


class BaseBlock:
    """
    blocks do not communicate with the report app directly, but instead return a dict
    """

    block = {}
    upload = False

    def __call__(self, report_client=None):
        self.postprocess(report_client)
        return self.block

    def postprocess(self, report_client=None):
        """
        code for postprocessing the block that involves the client, e.g. uploading images to the report app
        """
        pass

    def _paragraph(self, content, exclude_appblock=False):
        if exclude_appblock:
            return {
                "type": "paragraph",
                "content": self.preprocess_text_contents(content),
            }
        return {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {"type": "paragraph", "content": self.preprocess_text_contents(content)}
            ],
        }

    def preprocess_text_contents(self, content):
        """for processing text contents"""
        if isinstance(content, str):
            # allows for simple text input into paragraph
            return [{"type": "text", "text": content}]
        elif isinstance(content, (float, int, np.generic)):
            return [{"type": "text", "text": str(content)}]
        elif isinstance(content, dict):
            if content["type"] == "text":
                return [content]
            else:
                raise TypeError(f"This only accepts 'text' type, not {content['type']}")
        elif isinstance(content, list):
            content_list = []
            for c in content:
                processed_c = self.preprocess_text_contents(c)
                if not isinstance(processed_c, list):
                    processed_c = [processed_c]
                content_list += processed_c
            return content_list
        elif callable(content):
            return self.preprocess_text_contents(content())
        else:
            print(type(content))
            return [content]

    def postprocess_blocks(self, blocks, report_client=None):
        if isinstance(blocks, list):
            processed_blocks = []
            for b in blocks:
                processed_b = self.postprocess_block(b, report_client=report_client)
                processed_blocks.append(processed_b)
            return processed_blocks
        else:
            raise TypeError("'blocks' has to be of type list.")

    def postprocess_block(self, block, report_client=None):
        if isinstance(block, str):
            # if we input just a string it needs to output a paragraph block
            return self._paragraph(block, exclude_appblock=True)
        elif isinstance(block, (float, int, np.generic)):
            # to convert all edge case types to paragraph block
            return self._paragraph(str(block), exclude_appblock=True)
        elif isinstance(block, dict):
            # this is necessary to nest blocks that are already dictionaries within layout blocks
            input_block = block
            if block["type"] == "appBlock":
                input_block = block["content"][0]
            return input_block
        elif callable(block):
            return self.postprocess_block(
                block(report_client=report_client), report_client=report_client
            )
        else:
            print("this type got through", type(block))
            return block

    def _get_content_bytes(self, content):
        if isinstance(content, str):
            if "http" in content and "/":
                # online image
                content_bytes = io.BytesIO(requests.get(content).content).getvalue()
            else:
                # local filepath
                content_bytes = io.BytesIO(open(content, "rb").read()).getvalue()
        elif isinstance(content, bytes):
            content_bytes = content
        elif isinstance(content, io.BytesIO):
            content_bytes = content.getvalue()
        else:
            raise TypeError("'content' needs to be of type str, bytes or io.BytesIO.")
        return content_bytes


class BaseMark(BaseBlock):
    def preprocess_mark(self, content, mark, nested=False):
        """
        test cases:
        ItalicMark("test")
        ItalicMark(UnderlinMark("test"))
        ItalicMark([UnderlinMark("test"), "test2])
        [ItalicMark("test), "test2"]
        """
        if isinstance(content, str):
            # just handles simple plain text
            output = {"type": "text", "text": content, "marks": mark}
            return output if nested else [output]
        elif isinstance(content, (float, int, np.generic)):
            output = {"type": "text", "text": str(content), "marks": mark}
            return output if nested else [output]
        elif isinstance(content, dict):
            # this should append mark to an existing mark
            if "marks" in content:
                content["marks"] += mark
            return content
        elif isinstance(content, list):
            # this should output a list of {text, (marks)}
            content_list = []
            for c in content:
                processed_c = self.preprocess_mark(c, mark, nested=True)
                if not isinstance(processed_c, list):
                    processed_c = [processed_c]
                content_list += processed_c
            return content_list
        elif "blocks.marks" in str(type(content)):
            # this should call/activate a Mark, this check is kind of inefficient
            return self.preprocess_mark(content(), mark, nested=nested)
        else:
            output = {"type": "text", "text": content, "marks": mark}
            return output if nested else [output]



"""Blocks that can nest text blocks. They only accept content which is raw string, marks or a list of them."""
import numpy as np
from ai_dashboard.tabs.data_report.blocks.base import BaseBlock
from ai_dashboard.tabs.data_report.blocks.basic import Paragraph


class UnorderedList(BaseBlock):
    def _list_item(self, content, nested=False):
        # only takes a simple string and mark right now.
        return [
            {
                "type": "listItem",
                "content": [self._paragraph(content, exclude_appblock=True)],
            }
        ]

    def __init__(self, contents):
        if not isinstance(contents, list):
            raise TypeError("'contents' needs to be a List")
        output = []
        for c in contents:
            list_item = self._list_item(c)
            if not isinstance(list_item, list):
                list_item = [list_item]
            output += list_item
        self.block = {
            "type": "appBlock",
            "content": [{"type": "bulletList", "content": output}],
        }


class OrderedList(UnorderedList):
    def __init__(self, contents):
        if not isinstance(contents, list):
            raise TypeError("'contents' needs to be a List")
        output = []
        for c in contents:
            list_item = self._list_item(c)
            if not isinstance(list_item, list):
                list_item = [list_item]
            output += list_item
        self.block = {
            "type": "appBlock",
            "content": [{"type": "orderedList", "content": output}],
        }



from ai_dashboard.tabs.data_report.blocks.base import *
from ai_dashboard.tabs.data_report.blocks.basic import *
from ai_dashboard.tabs.data_report.blocks.lists import *
from ai_dashboard.tabs.data_report.blocks.connected import *
from ai_dashboard.tabs.data_report.blocks.media import *
from ai_dashboard.tabs.data_report.blocks.charts import *
from ai_dashboard.tabs.data_report.blocks.layout import *
from ai_dashboard.tabs.data_report.blocks.marks import *



import io
import uuid
import json
import numpy as np
import requests

from typing import Union, Optional, Dict, Any, List

from ai_dashboard.api import endpoints
from ai_dashboard.tabs import abstract_tab
from ai_dashboard.tabs.data_report import blocks
from ai_dashboard.tabs.data_report import constants
from ai_dashboard.tabs.data_report import markdown
from ai_dashboard.constants import Colours

import plotly.express as px


class DataReport(abstract_tab.Tab):
    ID = "REPORT"

    BLANK: Dict[str, Any] = {
        "activeFilterGroup": "",
        "color": None,
        "configuration": {
            "content": {
                "type": "doc",
                "content": [],
            }
        },
        "name": "",
        "type": ID,
    }

    def __init__(
        self,
        endpoints: endpoints.Endpoints,
        dataset_id: str,
        title: Optional[str] = None,
        name: Optional[str] = None,
        colour: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()

        self._endpoints = endpoints
        self.title = title

        if config is not None:
            self._config = config
        else:
            self.reset()

        self.dataset_id = dataset_id

        if colour is not None:
            colour_values = list(map(lambda x: x._value_, Colours))
            assert colour in colour_values, f"Colour must one of {colour_values}"

        self.config["colour"] = colour
        self.config["configuration"]["title"] = self.ID if title is None else title
        self.config["name"] = name

    def add_blocks(self, list_of_blocks):
        for block in list_of_blocks:
            output = block(self)
            self.config["configuration"]["content"]["content"].append(output)

    @classmethod
    def from_text(
        cls,
        text: str,
        title: str,
        name: Optional[str] = None,
        colour: Optional[str] = None,
    ):
        if colour is not None:
            assert colour in constants.colours
        data_report = cls(title=title)
        data_report.config = {
            "name": name,
            "type": "REPORT",
            "color": colour.upper() if colour is not None else colour,
            "configuration": {
                "title": title,
                "content": {
                    "type": "doc",
                    "content": [
                        {
                            "type": "appBlock",
                            "content": [
                                {
                                    "type": "paragraph",
                                    "attrs": {"textAlign": "left"},
                                    "content": [
                                        {
                                            "type": "text",
                                            "text": text,
                                        }
                                    ],
                                }
                            ],
                        }
                    ],
                },
            },
            "activeFilterGroup": "",
        }
        return data_report

    @classmethod
    def from_markdown(
        cls,
        text: str,
        title: str,
        name: Optional[str] = None,
        colour: Optional[str] = None,
    ):
        data_report = cls(title=title)
        md = markdown.MarkDown()
        content = md.parse(text)
        data_report.config = {
            "name": name,
            "type": cls.ID,
            "color": colour.upper() if colour is not None else colour,
            "configuration": {
                "title": title,
                "content": {
                    "type": "doc",
                    "content": content,
                },
            },
            "activeFilterGroup": "",
        }
        return data_report

    def add_markdown(self, text: str):
        md = markdown.MarkDown()
        content = md.parse(text)
        config = self.config["configuration"]
        if "content" not in config:
            self.config["configuration"]["content"] = {
                "type": "doc",
                "content": [],
            }
        self.config["configuration"]["content"]["content"] += content

    def add_connected_chart(
        self,
        groupby: List[Dict[str, Any]],
        metric: List[Dict[str, Any]],
        chart_type: str = "column",
        title: Optional[str] = None,
        page_size: int = 20,
        show_frequencies: bool = False,
        sentiment_field: str = "",
        y_axis_sort_field: str = "",
        sort_direction: str = "desc",
        filters: Optional[List[Dict[str, Any]]] = None,
        convert_filters: bool = True,
    ):
        block = blocks.ConnectedChart(
            groupby=groupby,
            metric=metric,
            chart_type=chart_type,
            title=title,
            page_size=page_size,
            show_frequencies=show_frequencies,
            sentiment_field=sentiment_field,
            y_axis_sort_field=y_axis_sort_field,
            sort_direction=sort_direction,
            filters=filters,
            convert_filters=convert_filters,
        )
        self.config["configuration"]["content"]["content"].append(block())

    add_aggregation = add_connected_chart

    def add_connected_metric(
        self,
        query: List[Dict[str, Any]],
        title: Optional[str] = None,
        show_frequencies: bool = False,
        sort_direction: str = "Descending",
    ):
        block = blocks.ConnectedMetric(
            query=query,
            title=title,
            show_frequencies=show_frequencies,
            sort_direction=sort_direction,
        )
        self.config["configuration"]["content"]["content"].append(block())

    add_metric = add_connected_metric

    def add_image(
        self, image_src: Union[bytes, str], title: str = "", width_percentage: int = 100
    ):
        block = blocks.Image(
            image_url=image_src, title=title, width_percentage=width_percentage
        )
        self.config["configuration"]["content"]["content"].append(block())

    def add_audio(
        self, audio_src: Union[bytes, str], title: str = "", width_percentage: int = 100
    ):
        block = blocks.Audio(
            audio_url=audio_src, title=title, width_percentage=width_percentage
        )
        self.config["configuration"]["content"]["content"].append(block())

    def add_video(
        self, video_src: Union[bytes, str], title: str = "", width_percentage: int = 100
    ):
        block = blocks.Video(
            video_url=video_src, title=title, width_percentage=width_percentage
        )
        self.config["configuration"]["content"]["content"].append(block())

    def add_plotly(
        self,
        fig: Any,
        title: str,
        width: int = None,
        height: int = None,
        width_percentage: int = 100,
        options=None,
    ):
        block = blocks.Plotly(
            fig=fig,
            title=title,
            width=width,
            height=height,
            width_percentage=width_percentage,
            options=options,
        )
        self.config["configuration"]["content"]["content"].append(block())

    def add_altair(
        self,
        fig: Any,
        title: str,
        width_percentage: int = 100,
    ):
        block = blocks.Altair(fig=fig, title=title, width_percentage=width_percentage)
        self.config["configuration"]["content"]["content"].append(block())

    def add_pyplot(
        self,
        fig,
        title: str = "",
        width: int = None,
        height: int = None,
        dpi: int = 150,
        add: bool = True,
        width_percentage: int = 100,
        **kwargs,
    ):
        block = blocks.Pyplot(
            fig=fig,
            title=title,
            width=width,
            height=height,
            dpi=dpi,
            add=add,
            width_percentage=width_percentage,
            **kwargs,
        )
        self.config["configuration"]["content"]["content"].append(block())
